/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

/*!
 *  Copyright (c) 2016 by Contributors
 * \file MxNetCpp.h
 * \brief meta include file for mxnet.cpp
 * \author Chuntao Hong, Zhang Chen
 */

#ifndef MXNET_CPP_MXNETCPP_H_
#define MXNET_CPP_MXNETCPP_H_

#include "mxnet-cpp/executor.hpp"
#include "mxnet-cpp/symbol.hpp"
#include "mxnet-cpp/ndarray.hpp"
#include "mxnet-cpp/monitor.hpp"
#include "mxnet-cpp/operator.hpp"
#include "mxnet-cpp/optimizer.hpp"
#include "mxnet-cpp/kvstore.hpp"
#include "mxnet-cpp/op.h"
#include "mxnet-cpp/op_suppl.h"
#include "mxnet-cpp/io.hpp"
#include "mxnet-cpp/metric.h"
#include "mxnet-cpp/initializer.h"

#endif  // MXNET_CPP_MXNETCPP_H_
